// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ==  http://www.amdis-fem.org                                              ==
// ==                                                                        ==
// ============================================================================
//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.



/** \file SecondOrderTerm.h */

#ifndef AMDIS_SECOND_ORDER_TERM_H
#define AMDIS_SECOND_ORDER_TERM_H

#include "AMDiS_fwd.h"
#include "OperatorTerm.h"
#include "ElInfo.h"
#include "AbstractFunction.h"

namespace AMDiS {

  using namespace std;

  /**
   * \ingroup Assembler
   * 
   * \brief
   * Describes the second order terms: \f$ \nabla \cdot A \nabla u(\vec{x}) \f$
   */
  class SecondOrderTerm : public OperatorTerm
  {
  public:
    /// Constructor.
    SecondOrderTerm(int deg) 
      : OperatorTerm(deg) 
    {}

    /// Destructor.
    virtual ~SecondOrderTerm() 
    {}

    /// Evaluation of \f$ \Lambda A \Lambda^t \f$ at all quadrature points.
    virtual void getLALt(const ElInfo *elInfo, 
			 vector<mtl::dense2D<double> > &result) const = 0;

    /// Evaluation of \f$ A \nabla u(\vec{x}) \f$ at all quadrature points.
    virtual void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
			  std::vector<WorldVector<double> > &result) = 0;

  protected:
    /// Evaluation of \f$ \Lambda \cdot A \cdot \Lambda^t\f$.
    void lalt(const DimVec<WorldVector<double> >& Lambda,
	      const WorldMatrix<double>& matrix,
	      mtl::dense2D<double>& LALt,
	      bool symm,
	      double factor) const;

    /** \brief
     * Evaluation of \f$ \Lambda \cdot A \cdot \Lambda^t\f$ for \f$ A \f$
     * the matrix having a ONE in the position \f$ (K,L) \f$
     * and ZEROS in all other positions.
     */
    static void lalt_kl(const DimVec<WorldVector<double> >& Lambda,
			int k, int l,
			mtl::dense2D<double>& LALt,
			double factor);

    /// Evaluation of \f$ \Lambda \cdot A \cdot \Lambda^t\f$ for A equal to the identity.
    inline void l1lt(const DimVec<WorldVector<double> >& Lambda, 
		     mtl::dense2D<double>& LALt,
		     double factor) const
    {
      const int dim = num_rows(LALt);
      
      for (int i = 0; i < dim; i++) {
	double val = 0.0;
	for (int k = 0; k < dimOfWorld; k++)
	  val += Lambda[i][k] * Lambda[i][k];
	val *= factor;
	LALt[i][i] += val;
	for (int j = i + 1; j < dim; j++) {
	  val = 0.0;
	  for (int k = 0; k < dimOfWorld; k++)
	    val += Lambda[i][k] * Lambda[j][k];
	  val *= factor;
	  LALt[i][j] += val;
	  LALt[j][i] += val;
	}    
      }      
    }

  };


  /**
   * \ingroup Assembler
   * 
   * \brief
   * Implements the laplace operator multiplied with a scalar factor:
   * \f$ f \cdot \Delta u(\vec{x}) \f$
   */
  class Simple_SOT : public SecondOrderTerm 
  {
  public:
    /// Constructor.
    Simple_SOT(double f = 1.0) 
      : SecondOrderTerm(0)  
    {
      factor = new double;
      *factor = f;

      setSymmetric(true);
    }

    /// Constructor.
    Simple_SOT(double *fptr) 
      : SecondOrderTerm(0), 
	factor(fptr)
    {
      setSymmetric(true);
    }

    /// Implements SecondOrderTerm::getLALt().
    inline void getLALt(const ElInfo *elInfo, vector<mtl::dense2D<double> > &LALt) const
    {
      const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();
      const int nPoints = static_cast<int>(LALt.size());

      for (int iq = 0; iq < nPoints; iq++) 
	l1lt(grdLambda, LALt[iq], (*factor));
    }

    /// Implenetation of SecondOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double f) 
    {
      int dow = Global::getGeo(WORLD);

      if (num_rows(D2UhAtQP) > 0) {
	for (int iq = 0; iq < nPoints; iq++) {
	  double resultQP = 0.0;
	  for (int i = 0; i < dow; i++) {
	    resultQP += D2UhAtQP[iq][i][i];
	  }
	  result[iq] += resultQP * f * (*factor);
	}
      }
    }

    /// Implenetation of SecondOrderTerm::weakEval().
    void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
		  std::vector<WorldVector<double> > &result) 
    {
      int nPoints = grdUhAtQP.size();
      for (int iq = 0; iq < nPoints; iq++)
	axpy(*factor, grdUhAtQP[iq], result[iq]);
    }

  private:
    /// Pointer to the factor.
    double *factor;
  };


  /**
   * \ingroup Assembler
   * 
   * \brief
   * SecondOrderTerm where A is a function which maps a DOFVector evaluated at
   * a given quadrature point to a WolrdMatrix:
   * \f$ \nabla \cdot A(v(\vec{x})) \nabla u(\vec{x}) \f$
   */
  class MatrixFct_SOT : public SecondOrderTerm 
  {
  public:
    /// Constructor.
    MatrixFct_SOT(DOFVectorBase<double> *dv, 
		  AbstractFunction<WorldMatrix<double>, double> *fct,
		  AbstractFunction<WorldVector<double>, WorldMatrix<double> > *div,
		  bool sym = false);

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		     Quadrature *quad = NULL);

    /// Implements SecondOrderTerm::getLALt().
    void getLALt(const ElInfo *elInfo, vector<mtl::dense2D<double> > &LALt) const;
  
    /// Implenetation of SecondOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);

    /// Implenetation of SecondOrderTerm::weakEval().
    void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
		  std::vector<WorldVector<double> > &result);

  protected:
    /// DOFVector to be evaluated at quadrature points.
    DOFVectorBase<double>* vec;

    /// Pointer to the values of the DOFVector at quadrature points.
    mtl::dense_vector<double> vecAtQPs;

    /// Function for A.
    AbstractFunction<WorldMatrix<double>, double>* matrixFct;

    ///
    AbstractFunction<WorldVector<double>, WorldMatrix<double> >* divFct;

    /// True, if \ref matrixFct produces always symmetric matrices.
    bool symmetric;
  };


  /** 
   * \ingroup Assembler
   *
   * \brief
   * SecondOrderTerm where A is a given fixed WorldMatrix<double>:
   * \f$ \nabla \cdot A \nabla u(\vec{x}) \f$
   */
  class Matrix_SOT : public SecondOrderTerm {
  public:
    /// Constructor
    Matrix_SOT(WorldMatrix<double> mat) 
      : SecondOrderTerm(0), matrix(mat)
    {
      symmetric = matrix.isSymmetric();
      setSymmetric(symmetric);
    }

    /// Implements SecondOrderTerm::getLALt().
    inline void getLALt(const ElInfo *elInfo, vector<mtl::dense2D<double> > &LALt) const
    {
      const DimVec<WorldVector<double> >& Lambda = elInfo->getGrdLambda();
      const int nPoints = static_cast<int>(LALt.size());

      for (int iq = 0; iq < nPoints; iq++) 
	lalt(Lambda, matrix, LALt[iq], symmetric, 1.0);
    }
  
    /// Implenetation of SecondOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);

    /// Implenetation of SecondOrderTerm::weakEval().
    void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
		  std::vector<WorldVector<double> > &result);

  protected:
    /// Matrix stroring A.
    WorldMatrix<double> matrix;

    /// True, if \ref matrix is symmetric.
    bool symmetric;
  };


  /** 
   * \ingroup Assembler
   *
   * \brief
   * SecondOrderTerm where A is a WorldMatrix<double> having a ONE in position IJ
   * and ZERO in all other positions
   * \f$ \nabla \cdot A \nabla u(\vec{x}) \f$
   */
  class FactorIJ_SOT : public SecondOrderTerm 
  {
  public:
    /// Constructor.
    FactorIJ_SOT(int x_i, int x_j, double f) 
      : SecondOrderTerm(0), xi(x_i), xj(x_j)
    {
      factor = new double;
      *factor = f;

      setSymmetric(xi == xj);
    }

    /// Constructor.
    FactorIJ_SOT(int x_i, int x_j, double *fptr) 
      : SecondOrderTerm(0), xi(x_i), xj(x_j), factor(fptr)
    {
      setSymmetric(xi == xj);
    }

    /// Implements SecondOrderTerm::getLALt().
    inline void getLALt(const ElInfo *elInfo, 
			vector<mtl::dense2D<double> > &LALt) const
    {
      const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();
      const int nPoints = static_cast<int>(LALt.size());

      for (int iq = 0; iq < nPoints; iq++)
	lalt_kl(grdLambda, xi, xj, LALt[iq], (*factor));
    }

    /// Implenetation of SecondOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double fac)
    {
      if (num_rows(D2UhAtQP) > 0)
	for (int iq = 0; iq < nPoints; iq++)
	  result[iq] += (*factor) * D2UhAtQP[iq][xi][xj] * fac;
    }

    /// Implenetation of SecondOrderTerm::weakEval().
    void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
		  std::vector<WorldVector<double> > &result) 
    {
      int nPoints = grdUhAtQP.size();
      for (int iq = 0; iq < nPoints; iq++)
	result[iq][xi] += (*factor) * grdUhAtQP[iq][xj];
    }

  private:
    /// Directions for the derivatives.
    int xi, xj;

    /// Pointer to the factor.
    double *factor;
  };


  /**
   * \ingroup Assembler
   *  
   * \brief
   * Laplace operator multiplied with a function evaluated at the quadrature
   * points of a given DOFVector:
   * \f$ f(v(\vec{x})) \Delta u(\vec{x}) \f$
   */
  class VecAtQP_SOT : public SecondOrderTerm 
  {
  public:
    /// Constructor.
    VecAtQP_SOT(DOFVectorBase<double> *dv, AbstractFunction<double, double> *af = NULL);

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		     Quadrature *quad = NULL);

    /// Implementation of \ref OperatorTerm::initElement() for multilpe meshes.
    void initElement(const ElInfo* smallElInfo,
		     const ElInfo* largeElInfo,
		     SubAssembler* subAssembler,
		     Quadrature *quad = NULL);

    /// Implements SecondOrderTerm::getLALt().
    void getLALt(const ElInfo *elInfo, vector<mtl::dense2D<double> > &LALt) const;    

    /// Implenetation of SecondOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);

    /// Implenetation of SecondOrderTerm::weakEval().
    void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
		  std::vector<WorldVector<double> > &result);

  protected:
    /// DOFVector to be evaluated at quadrature points.
    DOFVectorBase<double>* vec;

    /// Pointer to an array containing the DOFVector evaluated at quadrature points.
    mtl::dense_vector<double> vecAtQPs;

    /// Function evaluated at \ref vecAtQPs.
    AbstractFunction<double, double> *f;
  };

  /**
   * \ingroup Assembler
   *  
   * \brief
   * Laplace operator multiplied with a function evaluated at the quadrature
   * points of a given DOFVector:
   * \f$ f(v(\vec{x}), w(\vec{x})) \Delta u(\vec{x}) \f$
   */
  class Vec2AtQP_SOT : public SecondOrderTerm 
  {
  public:
    /// Constructor.
    Vec2AtQP_SOT(DOFVectorBase<double> *dv1,
		 DOFVectorBase<double> *dv2, 
		 BinaryAbstractFunction<double, double, double> *af = NULL);

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		     Quadrature *quad = NULL);

    /// Implements SecondOrderTerm::getLALt().
    void getLALt(const ElInfo *elInfo, vector<mtl::dense2D<double> > &LALt) const;    
    
    /// Implenetation of SecondOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);
    
    /// Implenetation of SecondOrderTerm::weakEval().
    void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
		  std::vector<WorldVector<double> > &result);
    
  protected:
    /// DOFVector to be evaluated at quadrature points.
    DOFVectorBase<double> *vec1, *vec2;
    
    /// Pointer to an array containing the DOFVector evaluated at quadrature points.
    mtl::dense_vector<double> vecAtQPs1, vecAtQPs2;
    
    /// Function evaluated at \ref vecAtQPs.
    BinaryAbstractFunction<double, double, double> *f;
  };


  /**
   * \ingroup Assembler
   * 
   * \brief
   * Laplace multiplied with a function evaluated at each quadrature point:
   * \f$ f(\vec{x}) \Delta u(\vec{x}) \f$
   */
  class CoordsAtQP_SOT : public SecondOrderTerm 
  {
  public:
    /// Constructor.
    CoordsAtQP_SOT(AbstractFunction<double, WorldVector<double> > *af)
      : SecondOrderTerm(af->getDegree()), g(af)
    {
      setSymmetric(true);
    }

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		     Quadrature *quad = NULL);

    /// Implements SecondOrderTerm::getLALt().
    void getLALt(const ElInfo *elInfo, vector<mtl::dense2D<double> > &LALt) const;    

    /// Implenetation of SecondOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);

    /// Implenetation of SecondOrderTerm::weakEval().
    void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
		  std::vector<WorldVector<double> > &result);


  protected:
    /// Stores coordinates at quadrature points. Set in \ref initElement().
    WorldVector<double>* coordsAtQPs;

    /// Function evaluated at quadrature points.
    AbstractFunction<double, WorldVector<double> > *g;
  };


  /**
   * \ingroup Assembler
   *
   * \brief
   * SecondOrderTerm where A is a function which maps the gradient of a 
   * DOFVector at each quadrature point to WorldMatrix<double>:
   * \f$ \nabla \cdot A(\nabla v(\vec{x})) \nabla u(\vec{x})\f$
   */
  class MatrixGradient_SOT : public SecondOrderTerm
  {
  public:
    /// Constructor.
    MatrixGradient_SOT(DOFVectorBase<double> *dv,
		       AbstractFunction<WorldMatrix<double>, WorldVector<double> > *af,
		       AbstractFunction<WorldVector<double>, WorldMatrix<double> > *divAf,
		       bool symm = false);

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		     Quadrature *quad = NULL);

    /// Implementation of \ref OperatorTerm::initElement() for multilpe meshes.
    void initElement(const ElInfo* smallElInfo, 
		     const ElInfo* largeElInfo,
		     SubAssembler* subAssembler,
		     Quadrature *quad = NULL);

    /// Implements SecondOrderTerm::getLALt().
    void getLALt(const ElInfo *elInfo, vector<mtl::dense2D<double> > &LALt) const;

    /// Implenetation of SecondOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);

    /// Implenetation of SecondOrderTerm::weakEval().
    void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
		  std::vector<WorldVector<double> > &result);

  protected:
    DOFVectorBase<double>* vec;

    /// Function which maps each entry in \ref gradAtQPs to a WorldMatrix<double>.
    AbstractFunction<WorldMatrix<double>, WorldVector<double> > *f;

    AbstractFunction<WorldVector<double>, WorldMatrix<double> > *divFct;

    /** \brief
     * Pointer to a WorldVector<double> array containing the gradients of the DOFVector
     * at each quadrature point.
     */
    mtl::dense_vector<WorldVector<double> > gradAtQPs;

    /// True, if \ref f provides always a symmetric WorldMatrix<double>.
    bool symmetric;

  private:
    /// Temporary matrix used in \ref MatrixGradient_SOT::weakEval.
    WorldMatrix<double> tmpMat;
  };


  /**
   * \ingroup Assembler
   *
   * \brief
   * Laplace multiplied with a function which maps the gradient of a DOFVector
   * at each quadrature point to a double:
   * \f$ f(\nabla v(\vec{x})) \Delta u(\vec{x}) \f$
   */
  class FctGradient_SOT : public SecondOrderTerm
  {
  public:
    /// Constructor.
    FctGradient_SOT(DOFVectorBase<double> *dv,  
		    AbstractFunction<double, WorldVector<double> > *af);

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		     Quadrature *quad = NULL);

    /// Implements SecondOrderTerm::getLALt().
    void getLALt(const ElInfo *elInfo, vector<mtl::dense2D<double> > &LALt) const;

    /// Implenetation of SecondOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);

    /// Implenetation of SecondOrderTerm::weakEval().
    void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
		  std::vector<WorldVector<double> > &result);

  protected:
    DOFVectorBase<double>* vec;

    /// Function wich maps \ref gradAtQPs to a double.
    AbstractFunction<double, WorldVector<double> > *f;

    /** \brief
     * Pointer to a WorldVector<double> array containing the gradients of the DOFVector
     * at each quadrature point.
     */
    mtl::dense_vector<WorldVector<double> > gradAtQPs;
  };


  /**
   * \ingroup Assembler
   *
   * \brief
   * Laplace multiplied with a function which maps the gradient of a DOFVector
   * at each quadrature point to a double:
   * \f$ \nabla \cdot A(v(\vec{x}), \nabla v(\vec{x})) \nabla u(\vec{x}) \f$
   */
  class VecMatrixGradientAtQP_SOT : public SecondOrderTerm
  {
  public:
    /// Constructor.
    VecMatrixGradientAtQP_SOT(DOFVectorBase<double> *dv,
			      BinaryAbstractFunction<WorldMatrix<double>, double, WorldVector<double> > *af,
			      AbstractFunction<WorldVector<double>, WorldMatrix<double> > *divAf,
			      bool symm = false);

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		     Quadrature *quad = NULL);

    /// Implements SecondOrderTerm::getLALt().
    void getLALt(const ElInfo *elInfo, vector<mtl::dense2D<double> > &LALt) const;

    /// Implenetation of SecondOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);

    /// Implenetation of SecondOrderTerm::weakEval().
    void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
		  std::vector<WorldVector<double> > &result);
    
  protected:
    DOFVectorBase<double>* vec;
    
    /// Function wich maps \ref gradAtQPs to a double.
    BinaryAbstractFunction<WorldMatrix<double>, 
			   double, WorldVector<double> > *f;
    
    AbstractFunction<WorldVector<double>, WorldMatrix<double> > *divFct;
    
    /** \brief
     * Pointer to a WorldVector<double> array containing the gradients of the DOFVector
     * at each quadrature point.
     */
    mtl::dense_vector<double> vecAtQPs;

    mtl::dense_vector<WorldVector<double> > gradAtQPs;
    
    bool symmetric;
  };


  /**
   * \ingroup Assembler
   *
   * \brief
   * Laplace multiplied with a function which maps the gradient of a DOFVector
   * at each quadrature point to a double:
   * \f$ f(\nabla v(\vec{x})) \Delta u(\vec{x}) \f$
   */
  class VecGradCoordsAtQP_SOT : public SecondOrderTerm
  {
  public:
    /// Constructor.
    VecGradCoordsAtQP_SOT(DOFVectorBase<double> *dv,
			  TertiaryAbstractFunction<double, double,
			  WorldVector<double>, WorldVector<double> > *af);

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		     Quadrature *quad = NULL);

    /// Implements SecondOrderTerm::getLALt().
    void getLALt(const ElInfo *elInfo, vector<mtl::dense2D<double> > &LALt) const;

    /// Implenetation of SecondOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);

    /// Implenetation of SecondOrderTerm::weakEval().
    void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
		  std::vector<WorldVector<double> > &result);

  protected:
    DOFVectorBase<double>* vec;

    /// Function wich maps \ref gradAtQPs to a double.
    TertiaryAbstractFunction<double, double, WorldVector<double>,
			     WorldVector<double> > *f;

    /** \brief
     * Pointer to a WorldVector<double> array containing the gradients of the DOFVector
     * at each quadrature point.
     */
    mtl::dense_vector<double> vecAtQPs;

    mtl::dense_vector<WorldVector<double> > gradAtQPs;

    WorldVector<double>* coordsAtQPs; 
  };


  /**
   * \ingroup Assembler
   *  
   * \brief
   * Laplace operator multiplied with a function evaluated at the quadrature
   * points of a given DOFVector:
   * \f$ -f(v(\vec{x})) \Delta u(\vec{x}) \f$
   */
  class VecAndCoordsAtQP_SOT : public SecondOrderTerm {
  public:
    /// Constructor.
    VecAndCoordsAtQP_SOT(DOFVectorBase<double> *dv, 
			 BinaryAbstractFunction<double, double, WorldVector<double> > *af);

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		     Quadrature *quad = NULL);

    /// Implements SecondOrderTerm::eval().
    void getLALt(const ElInfo *elInfo, vector<mtl::dense2D<double> > &LALt) const;    

    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);

    void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
		  std::vector<WorldVector<double> > &result);

  protected:
    /// DOFVector to be evaluated at quadrature points.
    DOFVectorBase<double>* vec;

    /// Pointer to an array containing the DOFVector evaluated at quadrature points.
    mtl::dense_vector<double> vecAtQPs;
  
    WorldVector<double>* coordsAtQPs;

    /// Function evaluated at \ref vecAtQPs.
    BinaryAbstractFunction<double, double,  WorldVector<double> > *f;
  };


  /**
   * \ingroup Assembler
   *
   * \brief
   * SecondOrderTerm where A is a function which maps the gradient of a 
   * DOFVector at each quadrature point to WorldMatrix<double>:
   * \f$ \nabla \cdot A(\nabla v(\vec{x})) \nabla u(\vec{x})\f$
   */
  class MatrixGradientAndCoords_SOT : public SecondOrderTerm
  {
  public:
    /// Constructor.
    MatrixGradientAndCoords_SOT(DOFVectorBase<double> *dv,
				BinaryAbstractFunction<WorldMatrix<double>,
				WorldVector<double> , 
				WorldVector<double> > *af,
				AbstractFunction<WorldVector<double>,
				WorldMatrix<double> > *divAf,
				bool symm = false);

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		     Quadrature *quad = NULL);

    /// Implements SecondOrderTerm::getLALt().
    void getLALt(const ElInfo *elInfo, vector<mtl::dense2D<double> > &LALt) const;

    /// Implenetation of SecondOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);

    /// Implenetation of SecondOrderTerm::weakEval().
    void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
		  std::vector<WorldVector<double> > &result);

  protected:
    DOFVectorBase<double>* vec;

    /// Function which maps each entry in \ref gradAtQPs to a WorldMatrix<double>.
    BinaryAbstractFunction<WorldMatrix<double>,
			   WorldVector<double>, WorldVector<double> > *f;

    AbstractFunction<WorldVector<double>, WorldMatrix<double> > *divFct;

    /** \brief
     * Pointer to a WorldVector<double> array containing the gradients of the DOFVector
     * at each quadrature point.
     */
    mtl::dense_vector<WorldVector<double> > gradAtQPs;

    WorldVector<double>* coordsAtQPs;

    /// True, if \ref f provides always a symmetric WorldMatrix<double>.
    bool symmetric;
  };


  class MatrixVec2_SOT : public SecondOrderTerm 
  {
  public:
    /// Constructor.
    MatrixVec2_SOT(DOFVectorBase<double> *dv1, DOFVectorBase<double> *dv2,  
		  BinaryAbstractFunction<double, double, double> *f,
		  WorldMatrix<double> Af,
		  bool sym = false);

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		     Quadrature *quad = NULL);

    /// Implements SecondOrderTerm::getLALt().
    void getLALt(const ElInfo *elInfo, vector<mtl::dense2D<double> > &LALt) const;
  
    /// Implenetation of SecondOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);

    /// Implenetation of SecondOrderTerm::weakEval().
    void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
		  std::vector<WorldVector<double> > &result);

  protected:
    /// DOFVector to be evaluated at quadrature points.
    DOFVectorBase<double> *vec1, *vec2;

    /// Pointer to the values of the DOFVector at quadrature points.
    mtl::dense_vector<double> vec1AtQPs, vec2AtQPs;

    /// Function for A.
    BinaryAbstractFunction<double, double, double> * fct;

    ///
    WorldMatrix<double> A;

    /// True, if \ref matrixFct produces always symmetric matrices.
    bool symmetric;
  };


  class General_SOT : public SecondOrderTerm
  {
  public:
    /// Constructor.
    General_SOT(std::vector<DOFVectorBase<double>*> vecs,
		std::vector<DOFVectorBase<double>*> grads,
		TertiaryAbstractFunction<WorldMatrix<double>, 
		  WorldVector<double>,
		  std::vector<double>, 
		  std::vector<WorldVector<double> > > *f,
		AbstractFunction<WorldVector<double>, 
		  WorldMatrix<double> > *divFct,
		bool symmetric);

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo*, SubAssembler*, Quadrature *quad= NULL);

    /// Implements SecondOrderTerm::getLALt().
    void getLALt(const ElInfo *elInfo, vector<mtl::dense2D<double> > &LALt) const;

    /// Implenetation of SecondOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);

    /// Implenetation of SecondOrderTerm::weakEval().
    void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
		  std::vector<WorldVector<double> > &result);

  protected:
    std::vector<DOFVectorBase<double>*> vecs_; 

    std::vector<DOFVectorBase<double>*> grads_;

    TertiaryAbstractFunction<WorldMatrix<double>, 
			     WorldVector<double>,
			     std::vector<double>, 
			     std::vector<WorldVector<double> > > *f_;

    AbstractFunction<WorldVector<double>, 
		     WorldMatrix<double> > *divFct_;

    WorldVector<double> *coordsAtQPs_;

    std::vector<mtl::dense_vector<double> > vecsAtQPs;

    std::vector<mtl::dense_vector<WorldVector<double> > > gradsAtQPs_;

    bool symmetric_;
  };


  class GeneralParametric_SOT : public SecondOrderTerm
  {
  public:
    /// Constructor.
    GeneralParametric_SOT(std::vector<DOFVectorBase<double>*> vecs,
			  std::vector<DOFVectorBase<double>*> grads,
			  QuartAbstractFunction<WorldMatrix<double>, 
			  WorldVector<double>,
			  WorldVector<double>,
			  std::vector<double>, 
			  std::vector<WorldVector<double> > > *f,
			  AbstractFunction<WorldVector<double>, 
			  WorldMatrix<double> > *divFct,
			  bool symmetric);

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo*, 
		     SubAssembler* ,
		     Quadrature *quad= NULL);

    /// Implements SecondOrderTerm::getLALt().
    void getLALt(const ElInfo *elInfo, vector<mtl::dense2D<double> > &LALt) const;

    /// Implenetation of SecondOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);

    /// Implenetation of SecondOrderTerm::weakEval().
    void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
		  std::vector<WorldVector<double> > &result);

  protected:
    std::vector<DOFVectorBase<double>*> vecs_; 

    std::vector<DOFVectorBase<double>*> grads_;

    QuartAbstractFunction<WorldMatrix<double>, 
			  WorldVector<double>,
			  WorldVector<double>,
			  std::vector<double>, 
			  std::vector<WorldVector<double> > > *f_;

    AbstractFunction<WorldVector<double>, 
		     WorldMatrix<double> > *divFct_;

    WorldVector<double> *coordsAtQPs_;
    WorldVector<double> elementNormal_;

    std::vector<mtl::dense_vector<double> > vecsAtQPs;

    std::vector<mtl::dense_vector<WorldVector<double> > > gradsAtQPs_;

    bool symmetric_;
  };


  /**
   * \ingroup Assembler
   *
   * \brief
   * Laplace multiplied with a function which maps the gradient of a DOFVector
   * at each quadrature point to a double:
   * \f$ f(\nabla v(\vec{x})) \Delta u(\vec{x}) \f$
   */
  class VecAndGradAtQP_SOT : public SecondOrderTerm
  {
  public:
    /// Constructor.
    VecAndGradAtQP_SOT(DOFVectorBase<double> *dv,
		       BinaryAbstractFunction<double, double, WorldVector<double> > *af);

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		     Quadrature *quad = NULL);


    /// Implements SecondOrderTerm::getLALt().
    void getLALt(const ElInfo *elInfo, vector<mtl::dense2D<double> > &LALt) const;

    /// Implements SecondOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);

    /// Implements SecondOrderTerm::weakEval().
    void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
		  std::vector<WorldVector<double> > &result);
  
  protected:
    /// DOFVector to be evaluated at quadrature points.
    DOFVectorBase<double>* vec;

    /// Vector v at quadrature points.
    mtl::dense_vector<double> vecAtQPs;

    /// Gradient at quadrature points.
    mtl::dense_vector<WorldVector<double> > gradAtQPs;

    /// Function for c.
    BinaryAbstractFunction<double, double, WorldVector<double> > *f;
  };


  /**
   * \ingroup Assembler
   *
   * \brief
   * Second order term: \f$ f(v(\vec{x}),\nabla w(\vec{x})) \Delta u(\vec{x})\f$
   */
  class VecGrad_SOT : public SecondOrderTerm
  {
  public:
    /// Constructor.
    VecGrad_SOT( DOFVectorBase<double> *dv1,
			  DOFVectorBase<double> *dv2,
		       BinaryAbstractFunction<double, double, WorldVector<double> > *f_)
      : SecondOrderTerm(f_->getDegree()), vec1(dv1), vec2(dv2), f(f_) 
    {}

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		     Quadrature *quad = NULL);

    /// Implements SecondOrderTerm::getLALt().
    void getLALt(const ElInfo *elInfo, vector<mtl::dense2D<double> > &LALt) const;

    /// Implements SecondOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);

    /// Implements SecondOrderTerm::weakEval().
    void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
		  std::vector<WorldVector<double> > &result);
  
  protected:
    /// DOFVectors to be evaluated at quadrature points.
    DOFVectorBase<double>* vec1;
    DOFVectorBase<double>* vec2;

    /// Vector v at quadrature points.
    mtl::dense_vector<double> vecAtQPs;

    /// Gradient at quadrature points.
    mtl::dense_vector<WorldVector<double> > gradAtQPs;

    /// Function for c.
    BinaryAbstractFunction<double, double, WorldVector<double> > *f;
  };


  /** 
   * \ingroup Assembler
   *
   * \brief
   * SecondOrderTerm where A(x) is a WorldMatrix having a in all positions
   * except possibly the position IJ
   * \f$ \nabla \cdot A(x) \nabla u(\vec{x}) \f$
   */
  class CoordsAtQP_IJ_SOT : public SecondOrderTerm 
  {
  public:
    /// Constructor.
    CoordsAtQP_IJ_SOT(AbstractFunction<double, WorldVector<double> > *af,
		      int x_i, int x_j) 
      : SecondOrderTerm(af->getDegree()), g(af), xi(x_i), xj(x_j)
    {
      setSymmetric(xi == xj);
    }

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		     Quadrature *quad = NULL);

    /// Implements SecondOrderTerm::getLALt().
    void getLALt(const ElInfo *elInfo, vector<mtl::dense2D<double> > &LALt) const;

    /// Implements SecondOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);

    /// Implements SecondOrderTerm::weakEval().
    void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
		  std::vector<WorldVector<double> > &result);
  
  private:
    /// Stores coordinates at quadrature points. Set in \ref initElement().
    WorldVector<double>* coordsAtQPs;

    /// Function evaluated at quadrature points.
    AbstractFunction<double, WorldVector<double> > *g;

    /// Directions for the derivatives.
    int xi, xj;
  };

  /**
   * \ingroup Assembler
   *  
   * \brief
   * SecondOrderTerm where A is a WorldMatrix having a in all positions
   * except possibly the position IJ, multiplied with a function
   * evaluated at the quadrature points of a given DOFVector:
   * \f$ \nabla \cdot f(v(\vec{x})) A \nabla u(\vec{x}) \f$
   */
  class VecAtQP_IJ_SOT : public SecondOrderTerm
  {
  public:
    /// Constructor.
    VecAtQP_IJ_SOT(DOFVectorBase<double> *dv,
		   AbstractFunction<double, double> *af,
		   int x_i, int x_j);

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		     Quadrature *quad = NULL);

    /// Implements SecondOrderTerm::getLALt().
    void getLALt(const ElInfo *elInfo, vector<mtl::dense2D<double> > &LALt) const;    

    /// Implements SecondOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);

    /// Implements SecondOrderTerm::weakEval().
    void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
		  std::vector<WorldVector<double> > &result); 
  
  protected:
    /// DOFVector to be evaluated at quadrature points.
    DOFVectorBase<double>* vec;

    /// Pointer to an array containing the DOFVector evaluated at quadrature points.
    mtl::dense_vector<double> vecAtQPs;

    /// Function evaluated at \ref vecAtQPs.
    AbstractFunction<double, double> *f;

  private:
    /// Directions for the derivatives.
    int xi, xj;
  };

  /**
   * \ingroup Assembler
   *
   * \brief
   * SecondOrderTerm where A is a WorldMatrix having a in all positions
   * except possibly the position IJ, multiplied with a function
   * evaluated at the quadrature points of two given DOFVectors:
   * \f$ \nabla \cdot f(v(\vec{x}),w(\vec{x})) A \nabla u(\vec{x}) \f$
   */
  class Vec2AtQP_IJ_SOT : public SecondOrderTerm
  {
  public:
    /// Constructor.
    Vec2AtQP_IJ_SOT(DOFVectorBase<double> *dv1,
		   DOFVectorBase<double> *dv2,
		   BinaryAbstractFunction<double, double, double> *af,
		   int x_i, int x_j);

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		     Quadrature *quad = NULL);

    /// Implements SecondOrderTerm::getLALt().
    void getLALt(const ElInfo *elInfo, vector<mtl::dense2D<double> > &LALt) const;

    /// Implements SecondOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);

    /// Implements SecondOrderTerm::weakEval().
    void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
		  std::vector<WorldVector<double> > &result);

  protected:
    /// DOFVector to be evaluated at quadrature points.
    DOFVectorBase<double>* vec1;
    DOFVectorBase<double>* vec2;

    /// Pointer to an array containing the DOFVector evaluated at quadrature points.
    mtl::dense_vector<double> vecAtQPs1;
    mtl::dense_vector<double> vecAtQPs2;

    /// Function evaluated at \ref vecAtQPs.
    BinaryAbstractFunction<double, double, double> *f;

  private:
    /// Directions for the derivatives.
    int xi, xj;
  };
}

#endif
