//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.


#include "Quadrature.h"
#include "FiniteElemSpace.h"
#include "ElInfo.h"

namespace AMDiS {

  template<typename T>
  void SubAssembler::getVectorAtQPs(DOFVectorBase<T>* vec, 
				    const ElInfo* elInfo,
				    Quadrature *quad,
				    mtl::dense_vector<T>& vecAtQPs)
  {
    FUNCNAME("SubAssembler::getVectorAtQPs()");

    TEST_EXIT_DBG(vec)("No DOF vector!\n");
    TEST_EXIT_DBG(elInfo->getMesh() == vec->getFeSpace()->getMesh())
      ("Vector and FE space do not fit together!\n");

    Quadrature *localQuad = quad ? quad : quadrature;
    vecAtQPs.change_dim(localQuad->getNumPoints());

    if (cachedValuesAtQPs[vec] && cachedValuesAtQPs[vec]->valid) {
      vecAtQPs = boost::any_cast<mtl::dense_vector<T>& >(cachedValuesAtQPs[vec]->values);
      return;
    }
    
    boost::any swap;
    if (!cachedValuesAtQPs[vec]) {
      swap = mtl::dense_vector<T>();
      cachedValuesAtQPs[vec] = new ValuesAtQPs(swap, false);
    }
    boost::any_cast<mtl::dense_vector<T>& >(cachedValuesAtQPs[vec]->values).change_dim(localQuad->getNumPoints());

    mtl::dense_vector<T>& values = boost::any_cast<mtl::dense_vector<T>& >(cachedValuesAtQPs[vec]->values);
    bool sameFeSpaces = 
      vec->getFeSpace() == rowFeSpace || vec->getFeSpace() == colFeSpace;

    if (opt && !quad && sameFeSpaces) {
      const BasisFunction *psi = rowFeSpace->getBasisFcts();
      const BasisFunction *phi = colFeSpace->getBasisFcts();
      if (vec->getFeSpace()->getBasisFcts() == psi)
	psiFast = updateFastQuadrature(psiFast, psi, INIT_PHI);
      else if(vec->getFeSpace()->getBasisFcts() == phi)
	phiFast = updateFastQuadrature(phiFast, phi, INIT_PHI);      
    }

    // calculate new values
    const BasisFunction *basFcts = vec->getFeSpace()->getBasisFcts();

    if (opt && !quad && sameFeSpaces) {
      if (psiFast->getBasisFunctions() == basFcts) {
	vec->getVecAtQPs(elInfo, NULL, psiFast, values);
      } else if (phiFast->getBasisFunctions() == basFcts) {
	vec->getVecAtQPs(elInfo, NULL, phiFast, values);
      } else {
	vec->getVecAtQPs(elInfo, localQuad, NULL, values);
      }
    } else {
      vec->getVecAtQPs(elInfo, localQuad, NULL, values);
    }

    cachedValuesAtQPs[vec]->valid = true;
    cachedValuesAtQPs[vec]->quad = localQuad;
    vecAtQPs = values;
  }


  template<typename T>
  void SubAssembler::getVectorAtQPs(DOFVectorBase<T>* vec, 
				    const ElInfo* smallElInfo,
				    const ElInfo* largeElInfo,
				    Quadrature *quad,
				    mtl::dense_vector<T>& vecAtQPs)
  {
    FUNCNAME("SubAssembler::getVectorAtQPs()");

    TEST_EXIT_DBG(vec)("No DOF vector!\n");
    Quadrature *localQuad = quad ? quad : quadrature;
    vecAtQPs.change_dim(localQuad->getNumPoints());

    if (cachedValuesAtQPs[vec] && cachedValuesAtQPs[vec]->valid) {
      vecAtQPs = boost::any_cast<mtl::dense_vector<T>& >(cachedValuesAtQPs[vec]->values);
      return;
    }    

    boost::any swap;
    if (!cachedValuesAtQPs[vec]) {
      swap = mtl::dense_vector<T>();
      cachedValuesAtQPs[vec] = new ValuesAtQPs(swap, false);
    }
    boost::any_cast<mtl::dense_vector<T>& >(cachedValuesAtQPs[vec]->values).change_dim(localQuad->getNumPoints());
    
    mtl::dense_vector<T>& values = boost::any_cast<mtl::dense_vector<T>& >(cachedValuesAtQPs[vec]->values);
    
    cachedValuesAtQPs[vec]->valid = true;
    vec->getVecAtQPs(smallElInfo, largeElInfo, localQuad, NULL, values);
    vecAtQPs = values;
  }


  template<typename T>
  void SubAssembler::getGradientsAtQPs( DOFVectorBase<T>* vec,
					const ElInfo* elInfo,
					Quadrature *quad,
					mtl::dense_vector<typename GradientType<T>::type>& grdAtQPs
				      )
  {
    FUNCNAME("SubAssembler::getGradientsAtQPs()");

    TEST_EXIT_DBG(vec)("No DOF vector!\n");
    Quadrature *localQuad = quad ? quad : quadrature;
    grdAtQPs.change_dim(localQuad->getNumPoints());

    if (cachedGradientsAtQPs[vec] && cachedGradientsAtQPs[vec]->valid) {
      grdAtQPs = boost::any_cast<mtl::dense_vector<typename GradientType<T>::type>& >(cachedGradientsAtQPs[vec]->values);
      return;
    }

    boost::any swap;
    if (!cachedGradientsAtQPs[vec]) {
      swap = mtl::dense_vector<typename GradientType<T>::type>();
      cachedGradientsAtQPs[vec] = new ValuesAtQPs(swap, false);
    }

    boost::any_cast<mtl::dense_vector<typename GradientType<T>::type>& >(cachedGradientsAtQPs[vec]->values).change_dim(localQuad->getNumPoints());

    mtl::dense_vector<typename GradientType<T>::type>& values = boost::any_cast<mtl::dense_vector<typename GradientType<T>::type>& >(cachedGradientsAtQPs[vec]->values);

    const BasisFunction *psi = rowFeSpace->getBasisFcts();
    const BasisFunction *phi = colFeSpace->getBasisFcts();

    bool sameFeSpaces =
      vec->getFeSpace() == rowFeSpace || vec->getFeSpace() == colFeSpace;

    if (opt && !quad && sameFeSpaces) {
      if (vec->getFeSpace()->getBasisFcts() == psi)
	psiFast = updateFastQuadrature(psiFast, psi, INIT_GRD_PHI);
      else if(vec->getFeSpace()->getBasisFcts() == phi)
	phiFast = updateFastQuadrature(phiFast, phi, INIT_GRD_PHI);
    }

    // calculate new values
    const BasisFunction *basFcts = vec->getFeSpace()->getBasisFcts();

    if (opt && !quad && sameFeSpaces) {
      if (psiFast->getBasisFunctions() == basFcts)
	vec->getGrdAtQPs(elInfo, NULL, psiFast, values);
      else if (phiFast->getBasisFunctions() == basFcts)
	vec->getGrdAtQPs(elInfo, NULL, phiFast, values);
      else
	vec->getGrdAtQPs(elInfo, NULL, phiFast, values);
    } else {
      vec->getGrdAtQPs(elInfo, localQuad, NULL, values);
    }

    cachedGradientsAtQPs[vec]->valid = true;
    grdAtQPs = values;
  }


  template<typename T>
  void SubAssembler::getGradientsAtQPs( DOFVectorBase<T>* vec,
					const ElInfo* smallElInfo,
					const ElInfo* largeElInfo,
					Quadrature *quad,
					mtl::dense_vector<typename GradientType<T>::type>& grdAtQPs)
  {
    FUNCNAME("SubAssembler::getGradientsAtQPs()");

    TEST_EXIT_DBG(vec)("No DOF vector!\n");

    Quadrature *localQuad = quad ? quad : quadrature;
    grdAtQPs.change_dim(localQuad->getNumPoints());

    if (cachedGradientsAtQPs[vec] && cachedGradientsAtQPs[vec]->valid) {
      grdAtQPs = boost::any_cast<mtl::dense_vector<typename GradientType<T>::type>& >(cachedGradientsAtQPs[vec]->values);
      return;
    }

    boost::any swap;
    if (!cachedGradientsAtQPs[vec]) {
      swap = mtl::dense_vector<typename GradientType<T>::type>();
      cachedGradientsAtQPs[vec] = new ValuesAtQPs(swap, false);
    }

    boost::any_cast<mtl::dense_vector<typename GradientType<T>::type>& >(cachedGradientsAtQPs[vec]->values).change_dim(localQuad->getNumPoints());

    mtl::dense_vector<typename GradientType<T>::type>& values = boost::any_cast<mtl::dense_vector<typename GradientType<T>::type>& >(cachedGradientsAtQPs[vec]->values);

    cachedGradientsAtQPs[vec]->valid = true;
    vec->getGrdAtQPs(smallElInfo, largeElInfo, localQuad, NULL, values);
    grdAtQPs = values;
  }

  template<typename T>
  void SubAssembler::getDerivativeAtQPs(DOFVectorBase<T>* vec,
					const ElInfo* elInfo,
					Quadrature *quad,
					int comp,
					mtl::dense_vector<T>& grdAtQPs)
  {
    FUNCNAME("SubAssembler::getGradientsAtQPs()");

    TEST_EXIT_DBG(vec)("No DOF vector!\n");

    Quadrature *localQuad = quad ? quad : quadrature;
    grdAtQPs.change_dim(localQuad->getNumPoints());

    if (cachedGradientsAtQPs[vec] && cachedGradientsAtQPs[vec]->valid) {      
      mtl::dense_vector<typename GradientType<T>::type> tmp = boost::any_cast<mtl::dense_vector<typename GradientType<T>::type>& >(cachedGradientsAtQPs[vec]->values);
      for (size_t iq = 0; iq < num_rows(tmp); iq++)
	grdAtQPs[iq] = tmp[iq][comp];
      return;
    }

    const BasisFunction *psi = rowFeSpace->getBasisFcts();
    const BasisFunction *phi = colFeSpace->getBasisFcts();

    bool sameFeSpaces =
      vec->getFeSpace() == rowFeSpace || vec->getFeSpace() == colFeSpace;

    if (opt && !quad && sameFeSpaces) {
      if (vec->getFeSpace()->getBasisFcts() == psi)
	psiFast = updateFastQuadrature(psiFast, psi, INIT_GRD_PHI);
      else if(vec->getFeSpace()->getBasisFcts() == phi)
	phiFast = updateFastQuadrature(phiFast, phi, INIT_GRD_PHI);
    }

    // calculate new values
    const BasisFunction *basFcts = vec->getFeSpace()->getBasisFcts();

    if (opt && !quad && sameFeSpaces) {
      if (psiFast->getBasisFunctions() == basFcts)
	vec->getDerivativeAtQPs(elInfo, NULL, psiFast, comp, grdAtQPs);
      else if (phiFast->getBasisFunctions() == basFcts)
	vec->getDerivativeAtQPs(elInfo, NULL, phiFast, comp, grdAtQPs);
      else
	vec->getDerivativeAtQPs(elInfo, NULL, phiFast, comp, grdAtQPs);
    } else {
      vec->getDerivativeAtQPs(elInfo, localQuad, NULL, comp, grdAtQPs);
    }
  }


  template<typename T>
  void SubAssembler::getDerivativeAtQPs(DOFVectorBase<T>* vec,
					const ElInfo* smallElInfo,
					const ElInfo* largeElInfo,
					Quadrature *quad,
					int comp,
					mtl::dense_vector<T>& grdAtQPs)
  {
    FUNCNAME("SubAssembler::getGradientsAtQPs()");

    TEST_EXIT_DBG(vec)("No DOF vector!\n");

    Quadrature *localQuad = quad ? quad : quadrature;
    grdAtQPs.change_dim(localQuad->getNumPoints());

    if (cachedGradientsAtQPs[vec] && cachedGradientsAtQPs[vec]->valid) {
      mtl::dense_vector<typename GradientType<T>::type> tmp = boost::any_cast<mtl::dense_vector<typename GradientType<T>::type>& >(cachedGradientsAtQPs[vec]->values);
      for (size_t iq = 0; iq < num_rows(tmp); iq++)
	grdAtQPs[iq] = tmp[iq][comp];
      return;
    }

    vec->getDerivativeAtQPs(smallElInfo, largeElInfo, localQuad, NULL, comp, grdAtQPs);
  }
  
}
