// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ==  http://www.amdis-fem.org                                              ==
// ==                                                                        ==
// ============================================================================
//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.



/** \file VtkWriter.h */

#ifndef AMDIS_VTKWRITER_H
#define AMDIS_VTKWRITER_H

#include <boost/iostreams/filtering_stream.hpp>
#include <boost/iostreams/device/file_descriptor.hpp>
#include <boost/iostreams/filter/gzip.hpp>
#include <boost/iostreams/filter/bzip2.hpp>

#include <fstream>
#include "BasisFunction.h"
#include "DataCollector.h"
#include "FileCompression.h"
#include "SystemVector.h"

namespace AMDiS {

  using namespace std;

  class VtkWriter
  {
  public:
    VtkWriter(vector<DataCollector<>*> *dc)
      : dataCollector(dc),
	compress(NONE)
    {
      degree = (*dataCollector)[0]->getFeSpace()->getBasisFcts()->getDegree();    
      dim = (*dataCollector)[0]->getMesh()->getDim();
    }  

    /// Writes a ParaView-VTK file.
    int writeFile(string name);

    /// Writes a pvtu file, which contains the links to all the rank files.
    static void writeParallelFile(string name, int nRanks,
				  string fnPrefix, string fnPostfix,
				  int nComponents);

    /// May be used to simply write ParaView files.
    static void writeFile(DOFVector<double> *values, 
			  string filename, 
			  bool writeParallel = true);

    /// May be used to simply write ParaView files.
    static void writeFile(DOFVector<double> &values, 
			  string filename,
			  bool writeParallel = true)
    {
      writeFile(&values, filename, writeParallel);
    }

    /// May be used to simply write ParaView files with a list of values.
    static void writeFile(vector<DOFVector<double>*> &values,
			  string filename, 
			  bool writeParallel = true);

    static void writeFile(WorldVector<DOFVector<double>*> &values,
			  string filename, 
			  bool writeParallel = true);

    static void writeFile(DOFVector<WorldVector<double> > *values,
			  string filename, 
			  bool writeParallel = true);

    static void writeFile(DOFVector<WorldVector<double> > &values,
			  string filename, 
			  bool writeParallel = true) 
    {
      writeFile(&values, filename, writeParallel);
    }

    static void writeFile(SystemVector *values,
			  string filename, 
			  bool writeParallel = true);

    static void writeFile(vector<DataCollector<>*> &dcList,
			  string filename,
			  bool writeParallel = true);

    /// Set a compressing method for file output.
    void setCompression(FileCompression c) 
    {
      compress = c;
    }

    /// Adds a new entry to a ParaView animation file.
    static int updateAnimationFile(AdaptInfo *adaptInfo,
				   string valueFilename,
				   vector<pair<double, string> > *paraViewAnimationFrames,
				   string animationFilename);
  protected:
    /// Writes the VTK file to an arbitrary stream.
    template<typename T>
    void writeFileToStream(T &file);

    /// Writes all coordinates of vertices and interpolation points to an 
    /// output file.
    template<typename T>
    void writeVertexCoords(T &file);

    /// Writes all values of vertices and interpolation point to an output file.
    template<typename T>
    void writeVertexValues(T &file, int componentNo);

    /// Writes the connectivity of all simplices to an output file.
    template<typename T>
    void writeConnectivity(T &file);

    /// Writes the connectivity for the case dim = 2 and degree = 2 to an 
    /// output file.
    template<typename T>
    void writeConnectivity_dim2_degree2(T &file);

    /// Writes the connectivity for the case dim = 2 and degree = 3 to an 
    /// output file.
    template<typename T>
    void writeConnectivity_dim2_degree3(T &file);

    /// Writes the connectivity for the case dim = 2 and degree = 4 to an 
    /// output file.
    template<typename T>
    void writeConnectivity_dim2_degree4(T &file);

    /// Writes a world coordinate to a given file.
    inline void writeCoord(boost::iostreams::filtering_ostream &file, 
			   WorldVector<double> coord) 
    {
      for (int i = 0; i < Global::getGeo(WORLD); i++)
	file << " " << scientific << coord[i];
      for (int i = Global::getGeo(WORLD); i < 3; i++)
	file << " "<<0.0;

      file << "\n";
    }

    /// Writes a world coordinate to a given file.
    inline void writeCoord(ofstream &file, WorldVector<double> coord) 
    {
      for (int i = 0; i < Global::getGeo(WORLD); i++)
	file << " " << scientific << coord[i];
      for (int i = Global::getGeo(WORLD); i < 3; i++)
	file << " "<<0.0;

      file << "\n";
    }

  private:
    /// List of DataCollectors, for each component of the problem one.
    vector<DataCollector<>*> *dataCollector;

    /// Defines if the file has to be compressed for ouput, and with which
    /// kind of compress method.
    FileCompression compress;

    /// Degree of the basis function of the problem.
    int degree;
    
    /// Dimension of the geometry.
    int dim;
  };
}

#include "VtkWriter.hh"

#endif
