// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ==  http://www.amdis-fem.org                                              ==
// ==                                                                        ==
// ============================================================================
//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.



/** \file MeshLevelData.h */

#include <iostream>
#include <set>
#include <vector>
#include <mpi.h>
#include "Global.h"

#ifndef AMDIS_MESH_LEVEL_DATA_H
#define AMDIS_MESH_LEVEL_DATA_H

namespace AMDiS {

  using namespace std;

  class MeshLevelData 
  {
  public:
    MeshLevelData();

    void init(std::set<int> &neighbourRanks);

    void addLevel(std::set<int> &ranksInDomain, int domainId);

    // Writes all data of this object to an output stream.
    void serialize(ostream &out);

    // Reads the object data from an input stream.
    void deserialize(istream &in);

    void print();

    std::set<int>& getLevelRanks(int level)
    {
      TEST_EXIT_DBG(level < nLevel)("Should not happen!\n");
      
      return levelRanks[level];
    }

    std::set<int>& getLevelNeighbours(int level)
    {
      TEST_EXIT_DBG(level < nLevel)("Should not happen!\n");
      
      return levelNeighbours[level];
    }

   int getLevelNumber()
   {
     return nLevel;
   }

   MPI::Intracomm& getMpiComm(int level)
   {
     TEST_EXIT_DBG(level < nLevel)("Should not happen!\n");

     return mpiComms[level];
   }

    MPI::Group& getMpiGroup(int level)
    {
      TEST_EXIT_DBG(level < nLevel)("Should not happen!\n");
      
      return mpiGroups[level];
    }

    int getLevelId(int level, int rank) 
    {
      if (level == 0)
	return rank;

      TEST_EXIT_DBG(level == 1 && rank <= 15)("Should not happen!\n");
      TEST_EXIT_DBG(MPI::COMM_WORLD.Get_size() == 16)("Should not happen!\n");

      if (rank == 0 || rank == 1 || rank == 2 || rank == 3)
	return 0;

      if (rank == 4 || rank == 5 || rank == 6 || rank == 7)
	return 1;
      
      if (rank == 8 || rank == 9 || rank == 10 || rank == 11)
	return 2;

      if (rank == 12 || rank == 13 || rank == 14 || rank == 15)
	return 3;

      ERROR_EXIT("Should not happen!\n");

      return -1;
    }

    int mapRank(int fromRank, int fromLevel, int toLevel)
    {
      int toRank = -1;

      MPI::Group::Translate_ranks(mpiGroups[fromLevel], 1, &fromRank, 
				  mpiGroups[toLevel], &toRank);

      TEST_EXIT_DBG(toRank != MPI::UNDEFINED)
	("Should not happen!\n");

      return toRank;			
    }

    bool rankInSubdomain(int rank, int level)
    {
      TEST_EXIT_DBG(level < nLevel)("Should not happen!\n");

      return static_cast<bool>(levelRanks[level].count(rank));
    }

  protected:
    int nLevel;

    vector<std::set<int> > levelRanks;
    
    vector<std::set<int> > levelNeighbours;

    vector<MPI::Intracomm> mpiComms;

    vector<MPI::Group> mpiGroups;
  };

}

#endif
