//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.


#include "parallel/ParallelDofMapping.h"
#include "parallel/StdMpi.h"

namespace AMDiS {

  using namespace std;


  void DofToMatIndex::getReverse(int rowIndex, int &component, int &dofIndex)
  {
    FUNCNAME("DofToMatIndex::getReverse()");

    for (map<int, map<DegreeOfFreedom, int> >::iterator it0 = data.begin();
	 it0 != data.end(); ++it0)
      for (map<DegreeOfFreedom, int>::iterator it1 = it0->second.begin();
	   it1 != it0->second.end(); ++it1)
	if (it1->second == rowIndex) {
	  component = it0->first;
	  dofIndex = it1->first;
	  return;
	}
  
    component = -1;
    dofIndex = -1;
  }


  void FeSpaceDofMap::clear()
  {
    dofMap.clear();

    nonRankDofs.clear();

    nRankDofs = 0;
    nLocalDofs = 0;
    nOverallDofs = 0;
    rStartDofs = 0;
  }


  void FeSpaceDofMap::update()
  {
    FUNCNAME("FeSpaceDofMap::update()");

    // === Compute local indices for all rank owned DOFs. ===
    
    for (DofMap::iterator it = dofMap.begin(); it != dofMap.end(); ++it)
      if (it->second.local == -1 && nonRankDofs.count(it->first) == 0)
	it->second.local = nRankDofs++;
    
    // === Compute number of local and global DOFs in the mapping. ===
    
    nOverallDofs = 0;
    rStartDofs = 0;
    mpi::getDofNumbering(mpiComm, nRankDofs, rStartDofs, nOverallDofs);
    
    // === If required, compute also the global indices. ===
    
    if (needGlobalMapping) {
      computeGlobalMapping();
      
      if (hasNonLocalDofs)
	computeNonLocalIndices();
    }
  }


  void FeSpaceDofMap::computeGlobalMapping()
  {
    FUNCNAME("FeSpaceDofMap::computeGlobalMapping()");

    for (DofMap::iterator it = dofMap.begin(); it != dofMap.end(); ++it)
      it->second.global = it->second.local + rStartDofs;
  }


  void FeSpaceDofMap::computeNonLocalIndices()
  {
    FUNCNAME("FeSpaceDofMap::computeNonLocalIndices()");

    TEST_EXIT_DBG(dofComm)("No DOF communicator defined!\n");

    typedef map<int, map<const FiniteElemSpace*, DofContainer> >::iterator it_type;

    // === Send all global indices of DOFs that are owned by the rank to all ===
    // === other ranks that also include this DOF.                           ===

    StdMpi<vector<int> > stdMpi(mpiComm);

    for (DofComm::Iterator it(dofComm->getSendDofs(), 0, feSpace); 
	 !it.end(); it.nextRank()) {
      int rank = it.getRank();
      if (meshLevel > 0)
	rank = levelData->mapRank(rank, 0, meshLevel);
      
      for (; !it.endDofIter(); it.nextDof())
	if (dofMap.count(it.getDofIndex()) && 
	    !nonRankDofs.count(it.getDofIndex()))
	  stdMpi.getSendData(rank).
	    push_back(dofMap[it.getDofIndex()].global);
    }

    stdMpi.updateSendDataSize();


    // === Check from which ranks this rank must receive some data. ===

    for (DofComm::Iterator it(dofComm->getRecvDofs(), 0, feSpace); 
	 !it.end(); it.nextRank()) {
      bool recvFromRank = false;
      for (; !it.endDofIter(); it.nextDof()) {
	if (nonRankDofs.count(it.getDofIndex())) {
	  recvFromRank = true;
	  break;
	}
      }

      if (recvFromRank) {
	int rank = it.getRank();
	if (meshLevel > 0)
	  rank = levelData->mapRank(rank, 0, meshLevel);

	stdMpi.recv(rank);
      }
    }


    // === Start communication to exchange global indices. ===

    stdMpi.startCommunication();


    // === And set the global indices for all DOFs that are not owned by rank. ===
    
    for (DofComm::Iterator it(dofComm->getRecvDofs(), 0, feSpace);
	 !it.end(); it.nextRank()) {
      int rank = it.getRank();
      if (meshLevel > 0)
	rank = levelData->mapRank(rank, 0, meshLevel);

      int i = 0;
      for (; !it.endDofIter(); it.nextDof())
	if (nonRankDofs.count(it.getDofIndex()))
	  dofMap[it.getDofIndex()].global = stdMpi.getRecvData(rank)[i++];
    }
  }


  void ParallelDofMapping::init(MeshLevelData &ldata,
				vector<const FiniteElemSpace*> &fe,
				vector<const FiniteElemSpace*> &uniqueFe,
				bool needGlobalMapping,
				bool bNonLocalDofs)
  {
    FUNCNAME("ParallelDofMapping::init()");

    levelData = &ldata;
    feSpaces = fe;
    feSpacesUnique = uniqueFe;
    hasNonLocalDofs = bNonLocalDofs;
    

    // === Init the mapping for all different FE spaces. ===

    for (vector<const FiniteElemSpace*>::iterator it = feSpacesUnique.begin();
	 it != feSpacesUnique.end(); ++it) {
      addFeSpace(*it);
      data[*it].setNeedGlobalMapping(needGlobalMapping);
      data[*it].setNonLocalDofs(hasNonLocalDofs);
    }
  }


  void ParallelDofMapping::clear()
  {
    FUNCNAME("ParallelDofMapping::clear()");

    TEST_EXIT_DBG(levelData)("No mesh level data object defined!\n");

    for (vector<const FiniteElemSpace*>::iterator it = feSpacesUnique.begin();
	 it != feSpacesUnique.end(); ++it)
      data[*it].clear();

    nRankDofs = -1;
    nLocalDofs = -1;
    nOverallDofs = -1;
    rStartDofs = -1;
    dofToMatIndex.clear();
  }

  
  void ParallelDofMapping::setMpiComm(MPI::Intracomm &m, int l)
  {
    mpiComm = m;
    meshLevel = l;

    for (vector<const FiniteElemSpace*>::iterator it = feSpacesUnique.begin();
	 it != feSpacesUnique.end(); ++it)
      data[*it].setMpiComm(m, l);
  }


  void ParallelDofMapping::setDofComm(DofComm &dc)
  {
    FUNCNAME("ParallelDofMapping::setDofComm()");

    dofComm = &dc;

    // Add the DOF communicator also to all FE space DOF mappings.
    for (vector<const FiniteElemSpace*>::iterator it = feSpacesUnique.begin();
	 it != feSpacesUnique.end(); ++it)
      data[*it].setDofComm(dc);
  }


  void ParallelDofMapping::addFeSpace(const FiniteElemSpace* feSpace)
  {
    FUNCNAME("ParallelDofMapping::addFeSpace()");
    
    if (data.count(feSpace))
      data.find(feSpace)->second.clear();
    else
      data.insert(make_pair(feSpace, FeSpaceDofMap(levelData)));
    
    data.find(feSpace)->second.setFeSpace(feSpace);
  }    


  int ParallelDofMapping::computeRankDofs()
  {
    FUNCNAME("ParallelDofMapping::computeRankDofs()");
    
    int result = 0;
    for (unsigned int i = 0; i < feSpaces.size(); i++) {
      TEST_EXIT_DBG(data.count(feSpaces[i]))("Should not happen!\n");
      result += data[feSpaces[i]].nRankDofs;
    }
    
    return result;
  }


  int ParallelDofMapping::computeLocalDofs()
  {
    FUNCNAME("ParallelDofMapping::computeLocalDofs()");
    
    int result = 0;
    for (unsigned int i = 0; i < feSpaces.size(); i++) {
      TEST_EXIT_DBG(data.count(feSpaces[i]))("Should not happen!\n");
      result += data[feSpaces[i]].nLocalDofs;
    }
    
    return result;
  }


  int ParallelDofMapping::computeOverallDofs()
  {
    FUNCNAME("ParallelDofMapping::computeOverallDofs()");

    int result = 0;
    for (unsigned int i = 0; i < feSpaces.size(); i++) {
      TEST_EXIT_DBG(data.count(feSpaces[i]))("Should not happen!\n");
      result += data.find(feSpaces[i])->second.nOverallDofs;
    }
    
    return result;
  }


  int ParallelDofMapping::computeStartDofs()
  {
    FUNCNAME("ParallelDofMapping::computeStartDofs()");

    int result = 0;
    for (unsigned int i = 0; i < feSpaces.size(); i++) {
      TEST_EXIT_DBG(data.count(feSpaces[i]))("Should not happen!\n");
      result += data.find(feSpaces[i])->second.rStartDofs;
    }
    
    return result;
  }


  void ParallelDofMapping::update()
  {
    FUNCNAME("ParallelDofMapping::update()");

    // First, update all FE space DOF mappings.
    for (vector<const FiniteElemSpace*>::iterator it = feSpacesUnique.begin();
	 it != feSpacesUnique.end(); ++it)
      data[*it].update();
    
    // Compute all numbers from this mappings.
    nRankDofs = computeRankDofs();
    nLocalDofs = computeLocalDofs();
    nOverallDofs = computeOverallDofs();
    rStartDofs = computeStartDofs();
    
    // And finally, compute the matrix indices.
    if (needMatIndex)
      computeMatIndex(needMatIndexFromGlobal);
  }
  

  void ParallelDofMapping::update(vector<const FiniteElemSpace*>& fe)
  {
    FUNCNAME("ParallelDofMapping::update()");

    for (vector<const FiniteElemSpace*>::iterator it = fe.begin();
	 it != fe.end(); ++it)
      if (find(feSpacesUnique.begin(), feSpacesUnique.end(), *it) == 
	  feSpacesUnique.end())
	ERROR_EXIT("Wrong FE space!\n");

    feSpaces = fe;
    update();
  }

  
  void ParallelDofMapping::computeMatIndex(bool globalIndex)
  {
    FUNCNAME("ParallelDofMapping::computeMatIndex()");

    dofToMatIndex.clear();
    
    // The offset is always added to the local matrix index. The offset for the
    // DOFs in the first FE spaces is the smalled global index of a DOF that is
    // owned by the rank.
    int offset = rStartDofs;

    // === Create the matrix indices for all component FE spaces. ===

    for (unsigned int i = 0; i < feSpaces.size(); i++) {

      // Traverse all DOFs of the FE space and create for all rank owned DOFs
      // a matrix index.
      DofMap& dofMap = data[feSpaces[i]].getMap();
      for (DofMap::iterator it = dofMap.begin(); it != dofMap.end(); ++it) {
	if (data[feSpaces[i]].isRankDof(it->first)) {
	  int globalMatIndex = it->second.local + offset;
	  if (globalIndex)
	    dofToMatIndex.add(i, it->second.global, globalMatIndex);
	  else
	    dofToMatIndex.add(i, it->first, globalMatIndex);
	}
      }

      // Increase the offset for the next FE space by the number of DOFs owned 
      // by the rank in the current FE space.
      offset += data[feSpaces[i]].nRankDofs;
	
      // If there are no non local DOFs, continue with the next FE space.
      if (!hasNonLocalDofs)
	continue;
      
      TEST_EXIT_DBG(dofComm != NULL)("No communicator given!\n");
      
      // === Communicate the matrix indices for all DOFs that are on some ===
      // === interior boundaries.                                         ===

      StdMpi<vector<DegreeOfFreedom> > stdMpi(mpiComm);
      for (DofComm::Iterator it(dofComm->getSendDofs(), 0, feSpaces[i]); 
	   !it.end(); it.nextRank()) {
	vector<DegreeOfFreedom> sendGlobalDofs;
	
	for (; !it.endDofIter(); it.nextDof())
	  if (dofMap.count(it.getDofIndex()))
	    if (globalIndex)
	      sendGlobalDofs.push_back(dofToMatIndex.get(i, dofMap[it.getDofIndex()].global));
	    else
	      sendGlobalDofs.push_back(dofToMatIndex.get(i, it.getDofIndex()));
	
	int rank = it.getRank();
	if (meshLevel > 0)
	  rank = levelData->mapRank(rank, 0, meshLevel);
	
	stdMpi.send(rank, sendGlobalDofs);
      }
      
      for (DofComm::Iterator it(dofComm->getRecvDofs(), 0, feSpaces[i]); 
	   !it.end(); it.nextRank()) {
	int rank = it.getRank();
	if (meshLevel > 0)
	  rank = levelData->mapRank(rank, 0, meshLevel);

	stdMpi.recv(rank);
      }

      stdMpi.startCommunication();
      
      for (DofComm::Iterator it(dofComm->getRecvDofs(), 0, feSpaces[i]); 
	   !it.end(); it.nextRank()) {
	int rank = it.getRank();
	if (meshLevel > 0)
	  rank = levelData->mapRank(rank, 0, meshLevel);
	
	int counter = 0;
	for (; !it.endDofIter(); it.nextDof()) {
	  if (dofMap.count(it.getDofIndex())) {
	    DegreeOfFreedom d = stdMpi.getRecvData(rank)[counter++];
	    if (globalIndex)
	      dofToMatIndex.add(i, dofMap[it.getDofIndex()].global, d);
	    else
	      dofToMatIndex.add(i, it.getDofIndex(), d);
	  }
	}
      }


      // === Communicate DOFs on periodic boundaries. ===

      stdMpi.clear();

      for (DofComm::Iterator it(dofComm->getPeriodicDofs(), 0, feSpaces[i]); 
	   !it.end(); it.nextRank()) {
	vector<DegreeOfFreedom> sendGlobalDofs;
	
	for (; !it.endDofIter(); it.nextDof())
	  if (dofMap.count(it.getDofIndex())) {
	    if (globalIndex) {
	      sendGlobalDofs.push_back(dofMap[it.getDofIndex()].global);
	      sendGlobalDofs.push_back(dofToMatIndex.get(i, dofMap[it.getDofIndex()].global));
	    } else {
	      sendGlobalDofs.push_back(dofToMatIndex.get(i, it.getDofIndex()));
	    }
	  }
	
	int rank = it.getRank();
	if (meshLevel > 0)
	  rank = levelData->mapRank(rank, 0, meshLevel);
	
	stdMpi.send(rank, sendGlobalDofs);
	stdMpi.recv(rank);
      }

      stdMpi.startCommunication();

      for (DofComm::Iterator it(dofComm->getPeriodicDofs(), 0, feSpaces[i]); 
	   !it.end(); it.nextRank()) {

	int rank = it.getRank();
	if (meshLevel > 0)
	  rank = levelData->mapRank(rank, 0, meshLevel);

	int counter = 0;

	for (; !it.endDofIter(); it.nextDof())
	  if (dofMap.count(it.getDofIndex())) {
	    if (globalIndex) {
	      TEST_EXIT_DBG(counter + 2 <= stdMpi.getRecvData(it.getRank()).size())
		("Should not happen!\n");

	      dofToMatIndex.add(i, 
				stdMpi.getRecvData(it.getRank())[counter],
				stdMpi.getRecvData(it.getRank())[counter + 1]);
	      counter += 2;
	    } else {
	      dofToMatIndex.add(i, it.getDofIndex(),
				stdMpi.getRecvData(it.getRank())[counter++]);
	    }
	  }
      }
    }
  }


  void ParallelDofMapping::createIndexSet(IS &is, 
					  int firstComponent, 
					  int nComponents)
  {
    FUNCNAME("ParallelDofMapping::createIndexSet()");

    TEST_EXIT_DBG(firstComponent + nComponents <= feSpaces.size())
      ("Should not happen!\n");

    TEST_EXIT_DBG(data.count(feSpaces[firstComponent]))
      ("No data for FE space at address %p!\n", feSpaces[firstComponent]);
    
    int firstRankDof = -1;
    FeSpaceDofMap &feMap = data.find(feSpaces[firstComponent])->second;
    DofMap &dofMap = feMap.getMap();
    for (DofMap::iterator it = dofMap.begin(); it != dofMap.end(); ++it) {
      if (feMap.isRankDof(it->first)) {
	if (needMatIndexFromGlobal)
	  firstRankDof = it->second.global;
	else
	  firstRankDof = it->first;
	break;
      }
    }
    
    TEST_EXIT_DBG(firstRankDof >= 0)("No rank DOF found!\n");    
    int firstMatIndex = dofToMatIndex.get(firstComponent, firstRankDof);

    int nRankRows = 0;
    for (int i = firstComponent; i < firstComponent + nComponents; i++)
      nRankRows += data.find(feSpaces[firstComponent])->second.nRankDofs;

    ISCreateStride(mpiComm, nRankRows, firstMatIndex, 1, &is);
  }

}
