#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_MODULE 0006
#define BOOST_TEST_NO_MAIN

#include <boost/test/unit_test.hpp>
#include <AMDiS.h>

using namespace AMDiS;
using namespace std;

/*
  - Test for interior boundaries of the subdomains on a 2-level 
     4x4 rank decomposition.
  - Compaires the interior boundary object with a serialization file.
  - Test for DOF communicators on the subdomains.
*/


void boundaryTest(RankToBoundMap &map0, RankToBoundMap &map1)
{
  BOOST_CHECK_MESSAGE(map0.size() == map1.size(), 
		      "Boundaries have different size!\n");

  for (RankToBoundMap::iterator it = map0.begin(); it != map0.end(); ++it) {
    if (map1.count(it->first) == 0) {
      MSG("Boundary with rank %d missing!\n", it->first);
      BOOST_ERROR("");
    }

    BOOST_CHECK_MESSAGE(it->second.size() == map1[it->first].size(), 
			"Wrong boundary size!\n");

    for (unsigned int i = 0; i < it->second.size(); i++)
      BOOST_CHECK_MESSAGE(it->second[i] == map1[it->first][i],
			  "Boundary is not equal!\n");
  }
}

BOOST_AUTO_TEST_CASE(amdis_mpi_feti_multilevel)
{
  BOOST_REQUIRE(MPI::COMM_WORLD.Get_size() == 16);

  ProblemStat ellipt("ellipt");
  ellipt.initialize(INIT_ALL);

  MeshDistributor *meshDist = MeshDistributor::globalMeshDistributor;
  meshDist->initParallelization();
  
  const FiniteElemSpace *feSpace = ellipt.getFeSpace();

  InteriorBoundary testBoundary;  
  string filename = "data/data0006/interior_boundary_p" + 
    lexical_cast<string>(MPI::COMM_WORLD.Get_rank()) + ".ser";

#if 0
  // Data was created with the following line
  ofstream myfile(filename.c_str());
  meshDist->getIntBoundarySd().serialize(myfile);
  myfile.close();
#else
  ifstream myfile(filename.c_str());
  testBoundary.deserialize(myfile, ellipt.getMesh());
  myfile.close();

  boundaryTest(testBoundary.getOwn(), meshDist->getIntBoundarySd().getOwn());
  boundaryTest(testBoundary.getOther(), meshDist->getIntBoundarySd().getOther());
  boundaryTest(testBoundary.getPeriodic(), meshDist->getIntBoundarySd().getPeriodic());
#endif


  Spreadsheet sheet;
  sheet.read("data/data0006/dofcomm_level_1.dat");
  vector<vector<double> > &data = sheet.getData();

  DofComm &dofCommSd = meshDist->getDofCommSd();
  DofComm::LevelDataType &sendDofsSd = dofCommSd.getSendDofs();
  DofComm::LevelDataType &recvDofsSd = dofCommSd.getRecvDofs();
  int nSendDofs = dofCommSd.getNumberDofs(sendDofsSd, 0, feSpace);
  int nRecvDofs = dofCommSd.getNumberDofs(recvDofsSd, 0, feSpace);

  int rank = MPI::COMM_WORLD.Get_rank();
  BOOST_CHECK(data[rank][0] == rank);
  BOOST_CHECK(data[rank][1] == nSendDofs);
  BOOST_CHECK(data[rank][2] == nRecvDofs);
}


int main(int argc, char **argv)
{
  AMDiS::init(argc, argv, "./init/test0006.dat.2d");

  boost::unit_test::unit_test_main(&init_unit_test, argc, argv);

  AMDiS::finalize();
}
